(in-package #:selenium)

(defclass iedoc-function ()
  ((name :initarg :name
	 :accessor iedoc-function-name)
   (parameters :initarg :parameters
	       :accessor iedoc-function-parameters)
   (comment :initarg :comment
	    :accessor iedoc-function-comment)))

(defmethod print-object ((iedoc-function iedoc-function) stream)
  (print-unreadable-object (iedoc-function stream :identity t :type t)
    (format stream "~A(~D)"
	    (iedoc-function-name iedoc-function)
	    (length (iedoc-function-parameters iedoc-function)))))

(defclass iedoc-parameter ()
  ((name :initarg :name
	 :accessor iedoc-parameter-name)
   (comment :initarg :comment
	    :accessor iedoc-parameter-comment)))

(defmethod print-object ((iedoc-parameter iedoc-parameter) stream)
  (print-unreadable-object (iedoc-parameter stream :identity t :type t)
    (format stream "~A" (iedoc-parameter-name iedoc-parameter))))

(defun make-iedoc-function (e)
  (let ((name (dom:get-attribute e "name")))
    (make-instance 'iedoc-function
		   :name name
		   :parameters (loop for p across (dom:get-elements-by-tag-name e "param")
				  collect (make-iedoc-parameter p))
		   :comment (make-iedoc-comment (elt (dom:get-elements-by-tag-name e "comment") 0)))))

(defun make-iedoc-parameter (e)
  (let ((name (dom:get-attribute e "name")))
    (make-instance 'iedoc-parameter
		   :name name
		   :comment (make-iedoc-comment e))))

(defun make-iedoc-comment (e)
  (labels ((make-iedoc-comment-r (e stream)
	     (loop for i across (dom:child-nodes e)
		do (if (dom:text-node-p i)
		       (write-string (dom:data i) stream)
		       (make-iedoc-comment-r i stream)))))
    (normalize-comment
     (with-output-to-string (stream)
       (make-iedoc-comment-r e stream)))))

(defun normalize-comment (comment)
  (cl-ppcre:regex-replace-all " {3,}" (cl-ppcre:regex-replace-all "[\\t\\n]" comment " ") " "))

(defun parse-iedoc (pathname)
  (let ((document (cxml:parse-file pathname (cxml-dom:make-dom-builder))))
    (loop for e across (dom:get-elements-by-tag-name document "function")
       collect (make-iedoc-function e))))

(defun convert-function-name (name)
  (intern (concatenate 'string "DO-" (string-upcase (cl-ppcre:regex-replace-all "([A-Z])" name "-\\1"))) "SELENIUM"))

(defun convert-parameter-name (name)
  (intern (string-upcase (cl-ppcre:regex-replace-all "([A-Z])" (cl-ppcre:regex-replace-all "ID" name "Id") "-\\1"))))

(defun convert-parameters (iedoc-function-parameters)
  (mapcar #'(lambda (parameter)
	      (convert-parameter-name (iedoc-parameter-name parameter)))
	  iedoc-function-parameters))

(defun replace-all (string part replacement &key (test #'char=))
  "Returns a new string in which all the occurences of the part 
  is replaced with replacement."
  (with-output-to-string (out)
    (loop with part-length = (length part)
       for old-pos = 0 then (+ pos part-length)
       for pos = (search part string
                         :start2 old-pos
                         :test test)
       do (write-string string out
                        :start old-pos
                        :end (or pos (length string)))
       when pos do (write-string replacement out)
       while pos)))

(defun encode (s)
  ;;; Must replace % first!
  (loop for disallowed-char in '("%" " " "<" ">" "#" "{" "}" "|" "\\" "^"
                                 "~" "[" "]" "`" ";" "/" "?" ":" "@" "=" "&"
                                 "$")
       do
       (setf s (replace-all s disallowed-char (format nil "%~x" (char-code (elt disallowed-char 0))))))
  s)

(defun marshall-request (command &rest parameters)
  (with-output-to-string (stream)
    (format stream "cmd=~A" command #+nil(html-encode:encode-for-http "asdf% asdf"))
    (dotimes (i (length parameters))
      (format stream "&~D=~A" (1+ i) (encode (elt parameters i))))))

(defun starts-with (s prefix)
  "Returns t if s starts with prefix"
  (let ((mismatch (mismatch s prefix
                            :test #'char=)))
    (or (null mismatch)
        (>= mismatch (length prefix)))))

(defun convert-result (s)
  (cond
    ((string= s "true") t)
    ((string= s "false") nil)
    (t s)))

(define-condition base-error (error)
  ())

(define-condition http-error (base-error)
  ((status-code :reader status-code
                :initarg :status-code)
   (reason :reader reason
           :initarg :reason))
  (:report (lambda (c s)
             (format s "HTTP Error(~d): ~A" 
                     (status-code c) 
                     (reason c)))))

(define-condition execution-error (base-error)
  ((description :reader description
                :initarg :description))
  (:report (lambda (c s)
             (format s "Selenium execution error: ~A" 
                     (description c)))))

(defun execute (url)
  (multiple-value-bind (reply status-code headers reply-from stream some-bool reason)
      (drakma:http-request url :method :get)
    (declare (ignore headers reply-from stream some-bool))
    (when (/= 200 status-code)
      (error 'http-error :status-code status-code :reason reason))
    (let ((results (split-sequence:split-sequence #\, reply)))
      (cond
        ((starts-with (first results) "OK") (convert-result (second results)))
        ((starts-with (first results) "ERROR:")
         (let ((err (second (split-sequence:split-sequence #\: (first results)))))
           (error 'execution-error :description err)))))))

(defun convert-function (iedoc-function)
  (let ((function-name (convert-function-name (iedoc-function-name iedoc-function)))
	(parameters (convert-parameters (iedoc-function-parameters iedoc-function))))
    `(defun ,function-name
	 (,@parameters &optional (session *selenium-session*))
       ,(iedoc-function-comment iedoc-function)
       (let ((query (concatenate 'string
                                 (marshall-request ,(iedoc-function-name iedoc-function) ,@parameters)
                                 (format nil "&sessionId=~d" session))))
         (execute (puri:merge-uris (make-instance 'puri:uri :query query) 
                                   *selenium-driver-url*))))))

#+nil (mapcar #'convert-function (parse-iedoc #p"/selenium-core-0.8.2/core/iedoc.xml"))

(defvar *selenium-driver-url* (puri:parse-uri "http://localhost:9999/selenium-server/driver/"))
(defvar *selenium-session* nil)
